{-----------------------------------------------------------------------------}
{ A form to provide gradient filled caption bars ala Microsoft Office.        }
{ Copyright 1996, Brad Stowers.  All Rights Reserved.                         }
{ This component can be freely used and distributed in commercial and private }
{ environments, provied this notice is not modified in any way and there is   }
{ no charge for it other than nomial handling fees.  Contact me directly for  }
{ modifications to this agreement.                                            }
{-----------------------------------------------------------------------------}
{ Feel free to contact me if you have any questions, comments or suggestions  }
{ at bstowers@pobox.com (preferred) or 72733,3374 on CompuServe.              }
{ The lateset version will always be available on the web at:                 }
{   http://www.pobox.com/~bstowers/delphi/                                    }
{-----------------------------------------------------------------------------}
{ Date last modified:  10/19/96                                               }
{-----------------------------------------------------------------------------}

{ ----------------------------------------------------------------------------}
{ TGradientForm v1.00                                                         }
{ ----------------------------------------------------------------------------}
{ Description:                                                                }
{   A form that paints it's caption bar in a gradient pattern, like the new   }
{   Microsoft Office applications.  It starts with black and moves gradually  }
{   to the system defined color.                                              }
{ Notes:                                                                      }
{   * The best way to use this form is to add it to your Object Repository.   }
{     Simply open this unit in Delphi, right click on the form and select Add }
{     To Repository.  Then, when you want a TGradientForm, you just select it }
{     from the repository (File | New) and use the "Inherit" option so you    }
{     don't have to see all this code in your form.                           }
{     If you have existing forms that you want converted to gradient forms,   }
{     simply add "GradForm" to your "Uses" clause, and change your form's     }
{     ancestor to TGradientForm.  An example:                                 }
{        Change:                                                              }
{           TMyForm = class(TForm)                                            }
{        To:                                                                  }
{           TMyForm = class(TGradientForm)                                    }
{   * Special thanks go to Michiel Ouwehand of Epic MegaGames for the         }
{     clipping region tips (See the WMNCPaint method) and for pointing out    }
{     the DrawFrameControl API function (see the PaintCaptionButtons method). }
{   * I've used strictly GDI calls for the painting in this component.  No    }
{     TCanvas, TBitmap, TBrush, etc.  This is because that although they are  }
{     extremely nice to use, they are not nearly as efficient.  That's not a  }
{     slam on them.  They have to be able to know how to do a lot of things,  }
{     and that requires overhead.  I have a very specific set of things to do }
{     here, and I am *very* interested in getting it to do it as fast as      }
{     possible, so I'm willing to sacrifice the convenience of the classes    }
{     for the speed of the API.  One day I'll sit down and do a speed         }
{     comparison to see if I really gained that much this way.  If you don't  }
{     understand the GDI calls, leave them alone, or use them to experiment   }
{     with in learning how to use them.                                       }
{   * This form will only work on Delphi 2.  I haven't used any calls that    }
{     would prevent it from working on Delphi 1, but there is a lot of stuff  }
{     that would have to be changed to make it look right.  For instance, I   }
{     paint an icon in the left corner for the system menu.  This is a Win95  }
{     style only.  Also, I allow for BorderStyles like bsToolWindow that      }
{     don't exist in Delphi 1.  Converting to Delphi 1 could be done, and     }
{     shouldn't be all that difficult since all the painting routines should  }
{     work fine.  I just don't do any Delphi 1 development any more, so I     }
{     leave it to one of you to implement.  If you do, I would appreciate it  }
{     if you could do it using $IFDEF WIN32, and send me the changes so       }
{     others can use it as well.                                              }
{   * As far as I know, there is no way to stop Windows from painting the     }
{     caption bar when it process the WM_NCACTIVATE message.  So, if you use  }
{     the GradientOnInactive property, it will cause a "flash" when the form  }
{     goes inactive.  We have to let Windows know that the form is inactive,  }
{     or it won't deactivate controls and such.  When it does that, it paints }
{     the caption in the inactive state.  If you don't use GradientOnInactive,}
{     then there's no problem.  If you do, then we have to paint over the     }
{     caption that Windows just painted.  This causes the "flash".  However,  }
{     because an upate region is passed in WM_NCPAINT, we can prevent Windows }
{     from painting the caption there.  This prevents the most obvious flash. }
{     If you know a way to prevent the default WM_NCACTIVATE handler from     }
{     painting the caption, please let me know.                               }
{                                                                             }
{ ----------------------------------------------------------------------------}
{ Revision History:                                                           }
{ 1.00:  + Initial release                                                    }
{ ----------------------------------------------------------------------------}

unit GradForm;

{$IFNDEF WIN32}
  ERROR!  This unit only available on Win32!
{$ENDIF}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Spin;

const
  MIN_GRADIENT_COLORS = 8;   // Any less than 8 colors doesn't look much like a gradient.
  MAX_GRADIENT_COLORS = 256; // Any more than 256 colors is not noticeable, and just
                             // slows the painting down.
  DEF_GRADIENT_COLORS = 64;  // This is a good compromise between speed and appearance.

type
  TGradientForm = class(TForm)
  private
    // Internal variables
    Colors: array[0..1, 0..MAX_GRADIENT_COLORS-1] of TColorRef;
    CaptionFont: HFONT;
    // Property variables
    FGradientColors: integer;
    FGradientOnInactive: boolean;

    // Internal methods
    procedure CalculateColors;
    function GetCaptionRect: TRect;
    function DrawCaption(FormDC: HDC; Active: boolean): TRect;
    procedure PaintMenuIcon(DC: HDC; var R: TRect);
    procedure FillRectGradient(DC: HDC; const R: TRect; Active: boolean);
    procedure PaintCaptionText(DC: HDC; R: TRect);
    procedure PaintCaptionButtons(DC: HDC; R: TRect);
    // Window message handlers
    procedure WMNCActivate(var Msg: TWMNCActivate); message WM_NCACTIVATE;
    procedure WMNCPaint(var Msg: TMessage); message WM_NCPAINT;
    procedure WMSysColorChange(var Msg: TWMSysColorChange); message WM_SYSCOLORCHANGE;
  protected
    // Property methods
    procedure SetGradientColors(Val: integer);
    procedure SetGradientOnInactive(Val: boolean);
  public
    // Overridden methods
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    // Properties
    // Determines the number of colors used to paint the gradient pattern.
    property GradientColors: integer
             read FGradientColors write SetGradientColors default DEF_GRADIENT_COLORS;
    //Determines if we should paint the gradient when the window is inactive.
    property GradientOnInactive: boolean
             read FGradientOnInactive write SetGradientOnInactive default FALSE;
  end;

var
  GradientForm: TGradientForm;

implementation

{$R *.DFM}

{$IFDEF TIMINGTEST}
var
  timetest: dword;
{$ENDIF}

// A variant record (known as a union in C) to allow easy access to the individual red,
// green, and blue values of a TColorRef (RGB) value.
type
  TRGBMap = packed record
    case boolean of
      TRUE:  ( RGBVal: DWORD );
      FALSE: ( Red,
               Green,
               Blue,
               Unused: byte );
  end;


constructor TGradientForm.Create(AOwner: TComponent);
var
  NCM: TNonClientMetrics;
begin
  // We set our new property values first so that they will be valid in the OnCreate
  // event handler.  The inherited Create is what calls that event, so we set up first.

  // Set the number of colors to use to create the gradient fill.
  FGradientColors := DEF_GRADIENT_COLORS;
  // Should we paint the gradient when window is inactive.
  FGradientOnInactive := FALSE;

  inherited Create(AOwner);

  // Calculate the colors we need to paint the gradient.
  CalculateColors;
  // Create a font for the caption bar.
  NCM.cbSize := SizeOf(NCM);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NCM, 0) then begin
    if BorderStyle in [bsToolWindow, bsSizeToolWin] then
      CaptionFont := CreateFontIndirect(NCM.lfSmCaptionFont)
    else
      CaptionFont := CreateFontIndirect(NCM.lfCaptionFont);
  end else
    CaptionFont := 0;
end;

destructor TGradientForm.Destroy;
begin
  // Clean up the font we created.
  if CaptionFont <> 0 then
    DeleteObject(CaptionFont);
  inherited Destroy;
end;

procedure TGradientForm.SetGradientColors(Val: integer);
begin
  if (Val = FGradientColors) or
     (Val < MIN_GRADIENT_COLORS) or (Val > MAX_GRADIENT_COLORS) then exit;
  FGradientColors := Val;
  // The number of colors have changes, we need to recalculate the colors we use to paint.
  CalculateColors;
  // Make the non client area repaint.
  SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_DRAWFRAME or SWP_NOACTIVATE or SWP_NOMOVE or
                                      SWP_NOSIZE or SWP_NOZORDER);
end;

procedure TGradientForm.SetGradientOnInactive(Val: boolean);
begin
  if FGradientOnInactive = Val then exit;
  FGradientOnInactive := Val;
  if GetActiveWindow <> Handle then
    // Make the non client area repaint.
    SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_DRAWFRAME or SWP_NOACTIVATE or SWP_NOMOVE or
                                        SWP_NOSIZE or SWP_NOZORDER);
end;

procedure TGradientForm.CalculateColors;
var
  SysColor: TRGBMap;
  RedPct,
  GreenPct,
  BluePct: real;
  x,
  Band: byte;
begin
  // Get colors for both active and inactive captions.
  for x := 0 to 1 do begin
    if x = 0 then
      SysColor.RGBVal := GetSysColor(COLOR_INACTIVECAPTION)
    else
      SysColor.RGBVal := GetSysColor(COLOR_ACTIVECAPTION);
    // Figure out the percentage of each RGB value needed for banding
    with SysColor do begin
      RedPct   := Red / (FGradientColors-1);
      GreenPct := Green / (FGradientColors-1);
      BluePct  := Blue / (FGradientColors-1);
    end;
    // Use the percentage of each color to create each band color.
    for Band := 0 to (FGradientColors-1) do
      Colors[x][Band] := RGB(round(RedPct * (Band)),
                             round(GreenPct * (Band)), round(BluePct * (Band)));
  end;
end;

// The caption rect is the rectangle we are interested in painting.  This will be the
// area that contains the caption icon, text and buttons.
function TGradientForm.GetCaptionRect: TRect;
begin
  // if we have no border style, then just set the rectange empty.
  if BorderStyle = bsNone then
    SetRectEmpty(Result)
  else begin
    GetWindowRect(Handle, Result);
    // Convert rect from screen (absolute) to client (0 based) coordinates.
    OffsetRect(Result, -Result.Left, -Result.Top);
    // Shrink rectangle to allow for window border.  We let Windows paint it.
    case BorderStyle of
      bsToolWindow, bsSingle, bsDialog:
          InflateRect(Result, -GetSystemMetrics(SM_CXFIXEDFRAME),
                              -GetSystemMetrics(SM_CYFIXEDFRAME));
      bsSizeable, bsSizeToolWin:
          InflateRect(Result, -GetSystemMetrics(SM_CXSIZEFRAME),
                              -GetSystemMetrics(SM_CYSIZEFRAME));
    end;

    // Set the appropriate height of caption bar.
    if BorderStyle in [bsToolWindow, bsSizeToolWin] then
      Result.Bottom := Result.Top + GetSystemMetrics(SM_CYSMCAPTION) - 1
    else
      Result.Bottom := Result.Top + GetSystemMetrics(SM_CYCAPTION) - 1;
  end;
end;

// Paint the icon for the system menu.
procedure TGradientForm.PaintMenuIcon(DC: HDC; var R: TRect);
var
  IconHandle: HIcon;
  IW, IH,
  Size: integer;
  TempDC: HDC;
  OldBrush: HBrush;
  OldBmp,
  TempBMP: HBitmap;
begin
  // Does the form have an icon assigned to it?
  if Icon.Handle <> 0 then
    IconHandle := Icon.Handle
  // If not, does the application have an icon?
  else if Application.Icon.Handle <> 0 then
    IconHandle := Application.Icon.Handle
  // If not, then just use the system defined application icon.
  else
    IconHandle := LoadIcon(0, IDI_APPLICATION);

  with R do begin
    Size := Bottom-Top-2;
    // Create a temporary device context to paint the icon on.
    TempDC := CreateCompatibleDC(DC);
    // How big should the icon be?
    IW := GetSystemMetrics(SM_CXICON);
    IH := GetSystemMetrics(SM_CYICON);
    // Create a temporary bitmap for painting the icon in.
    TempBMP := CreateCompatibleBitmap(DC, IW, IH);
    OldBmp := SelectObject(TempDC, TempBMP);
    OldBrush := SelectObject(TempDC, GetStockObject(BLACK_BRUSH));
    try
      // Paint the background black
      PatBlt(TempDC, 0, 0, IW, IH, PATCOPY);
      // Draw the icon
      DrawIcon(TempDC, 0, 0, IconHandle);
      // Stretch the icon down to the size we need and copy it to the real device context.
      StretchBlt(DC, Left, Top+1, Size, Size, TempDC, 0, 0, IW, IH, SRCCOPY);
    finally
      // clean up all the temporary drawing objects.
      SelectObject(TempDC, OldBrush);
      SelectObject(TempDC, OldBmp);
      DeleteObject(TempBMP);
      DeleteDC(TempDC);
    end;
  end;
  Inc(R.Left, Size+1);
end;

// Paint the given rectangle with the gradient pattern.
procedure TGradientForm.FillRectGradient(DC: HDC; const R: TRect; Active: boolean);
var
  OldBrush,
  Brush: HBrush;
  Step: real;
  Band: integer;
  H: integer;
begin
  // Determine how large each band should be in order to cover the
  // rectangle (one band for every color intensity level).
  Step := (R.Right - R.Left) / FGradientColors;
  H := R.Bottom - R.Top;
  // Start filling bands
  for Band := 0 to (FGradientColors-1) do begin
    // Create a brush with the appropriate color for this band
    Brush := CreateSolidBrush(Colors[ord(Active)][Band]);
    // Select that brush into the temporary DC.
    OldBrush := SelectObject(DC, Brush);
    try
      // Fill the rectangle using the selected brush -- PatBlt is faster than FillRect
      PatBlt(DC, round(Band*Step), 0, round((Band+1)*Step), H, PATCOPY);
    finally
      // Clean up the brush
      SelectObject(DC, OldBrush);
      DeleteObject(Brush);
    end;
  end; // for
end;

procedure TGradientForm.PaintCaptionText(DC: HDC; R: TRect);
var
  OldColor: TColorRef;
  OldMode: integer;
  OldFont: HFont;
begin
  Inc(R.Left, 2);
  // Set the text color to white, not the system color.  We always color from black to
  // the system color, so we will be almost guaranteed of having a nearly black background
  // under the text.  White is the safest color to use in this case (Microsoft thinks so
  // too, so I can't be too wrong).
  OldColor := SetTextColor(DC, RGB(255,255,255));
  // Set the background text painting mode to transparent so that drawing text doesn't
  // distrub the gradient we just painted.  If you didn't do this, then drawing text would
  // also fill the text rectangle with a solid background color, screwing up our gradient.
  OldMode := SetBkMode(DC, TRANSPARENT);
  // Select in the system defined caption font (see Create constructor).
  if CaptionFont <> 0 then
    OldFont := SelectObject(DC, CaptionFont)
  else
    OldFont := 0;
  try
    // Draw the text making it left aligned, centered vertically, allowing no line breaks.
    DrawText(DC, PChar(Caption), -1, R, DT_LEFT or DT_VCENTER or DT_SINGLELINE);
  finally
    // Clean up all the drawing objects.
    if OldFont <> 0 then
      SelectObject(DC, OldFont);
    SetBkMode(DC, OldMode);
    SetTextColor(DC, OldColor);
  end;
end;

// Paint the min/max/help/close buttons.
procedure TGradientForm.PaintCaptionButtons(DC: HDC; R: TRect);
var
  BtnWidth: integer;
  Flag: UINT;
begin
  InflateRect(R, -2, -2);
  if BorderStyle in [bsToolWindow, bsSizeToolWin] then begin
    // Tool windows only have the close button, nothing else.
    R.Left := R.Right - GetSystemMetrics(SM_CXSMSIZE) + 2;
    DrawFrameControl(DC, R, DFC_CAPTION, DFCS_CAPTIONCLOSE);
  end else begin
    BtnWidth := GetSystemMetrics(SM_CXSMSIZE);
    R.Left := R.Right - BtnWidth - 2;
    // if it has system menu, it has a close button.
    if biSystemMenu in BorderIcons then begin
      DrawFrameControl(DC, R, DFC_CAPTION, DFCS_CAPTIONCLOSE);
      OffsetRect(R, -BtnWidth-4, 0);
    end;
    // Minimize and Maximized don't show up at all if BorderStyle is bsDialog
    if BorderStyle <> bsDialog then begin
      Flag := DFCS_CAPTIONMAX;
      // if it doesn't have max in style, then it shows up disabled
      if not (biMaximize in BorderIcons) then
        Flag := Flag or DFCS_INACTIVE;
      DrawFrameControl(DC, R, DFC_CAPTION, Flag);
      OffsetRect(R, -BtnWidth-2, 0);

      Flag := DFCS_CAPTIONMIN;
      // if it doesn't have min in style, then it shows up disabled
      if not (biMinimize in BorderIcons) then
        Flag := Flag or DFCS_INACTIVE;
      DrawFrameControl(DC, R, DFC_CAPTION, Flag);
      OffsetRect(R, -BtnWidth-2, 0);
    end else // Help only shows up in bsDialog style
      if (biHelp in BorderIcons) then
        DrawFrameControl(DC, R, DFC_CAPTION, DFCS_CAPTIONHELP);
  end;
end;

// This draws everything we need, using all the other methods to put it all together.
function TGradientForm.DrawCaption(FormDC: HDC; Active: boolean): TRect;
var
  R: TRect;
  OldBmp,
  Bmp: HBitmap;
  BmpDC: HDC;
  w,h:integer;
begin
  // Get only the portion we need to draw.
  R := GetCaptionRect;
  Result := R;

  // Convert to logical (0-based) coordinates
  OffsetRect(R, -R.Left, -R.Top);

  W := R.Right - R.Left;
  H := R.Bottom - R.Top;

  // Create a temporary device context to draw on.  Drawing on a temporary DC and copying
  // it to the real DC accomplishes two things:
  // 1) It is faster because Windows doesn't have to draw anything in the temporary DC on
  // the screen, it only draws when you paint something on a real DC.  Then it just draws
  // everything at once when we copy it, instead of drawing a little, do some
  // calculations, draw a little, etc.
  // 2) It looks much better because it is drawn faster.  It reduces the "flicker" that
  // you would see from each individual part being drawn, especially the gradient bands.
  BmpDC := CreateCompatibleDC(FormDC);
  Bmp := CreateCompatibleBitmap(FormDC, W, H);
  OldBmp := SelectObject(BmpDC, Bmp);

  try
    // Draw the gradient background in the temporary DC.
    FillRectGradient(BmpDC, R, Active);
    Inc(R.Left, 1);
    // Do we need to paint an icon for the system menu?
    if (biSystemMenu in BorderIcons) and (BorderStyle in [bsSingle, bsSizeable]) then
      PaintMenuIcon(BmpDC, R); // PaintMenuIcon will adjust the rect so that future
                               // drawing operations happen in the right spot.
    PaintCaptionText(BmpDC, R); // Paint the caption text.
    PaintCaptionButtons(BmpDC, R); // Paint the min/max/help/close buttons.

    // Copy the gradient caption bar to the real DC.
    BitBlt(FormDC, Result.Left, Result.Top, W, H, BmpDC, 0, 0, SRCCOPY);
  finally
    // Clean up all the temporary drawing objects.
    SelectObject(BmpDC, OldBmp);
    DeleteObject(Bmp);
    DeleteDC(BmpDC);
  end;
end;

// Windows sends this message when the window has been activated or deactivated.
procedure TGradientForm.WMNCActivate(var Msg: TWMNCActivate);
var
  DC: HDC;
begin
  // We don't have an update region to work with here, so we can't exclude our painting
  // area from what is painted by default.  That means that we have to let the default
  // processing paint the entire bar, and then we paint over it.  It produces a nasty
  // flicker, so I don't much like it.  If you figure out a way around it, please let
  // me know.
  inherited;

  if Msg.Active or FGradientOnInactive then begin
    // Get the DC we need to paint in.  GetDC would only get the DC for the client area, we
    // need it for non-client area, too, so we use GetWindowDC.
    DC := GetWindowDC(Handle);
    try
      DrawCaption(DC, Msg.Active);
    finally
      ReleaseDC(Handle, DC); // NEVER leave this hanging.
    end;
  end;
end;

// Windows sends this message whenever any part of the non-client area (caption, window
// border) needs repainting.
procedure TGradientForm.WMNCPaint(var Msg: TMessage);
var
  WR,
  R:TRect;
  DC: HDC;
  MyRgn: HRGN;
  DeleteRgn: boolean;
begin
  DeleteRgn := FALSE;
  // The region that needs painting is passed in WParam.  A region is a Windows object
  // used to describe the non-rectangular area used by a combination of rectangles.
  MyRgn := Msg.wParam;
  DC := GetWindowDC(Handle);
  try
    GetWindowRect(Handle, WR);
    // Select the update region as the clipping region.  Clipping regions guarantee that
    // any painting done outside of the selected region is not shown (thrown away).
    if SelectClipRgn(DC, MyRgn) = ERROR then begin
      // We got passed an invalid region.  Generally, this happens when the window is
      // first created.  We'll create our own region (the rectangle that makes up the
      // entire window) and use that instead.
      with WR do
        MyRgn := CreateRectRgn(Left, Top, Right, Bottom);
      SelectClipRgn(DC, MyRgn);
      DeleteRgn := TRUE;
    end;
    // Convert the clipping region coordinates from screen to client.
    OffsetClipRgn(DC, -WR.Left, -WR.Top);
    // Draw our gradient caption.
    R := DrawCaption(DC, GetActiveWindow=Handle);
    // Here's the trick.  DrawCaption returns the rectangle that we painted.  We now
    // exclude that rectangle from the clipping region.  This guarantees that any further
    // painting that occurs will not happen in this rectangle.  That means that when we
    // let the default painting for WM_NCPAINT occur, it will not paint over our gradient.
    // It only paints the stuff that we didn't, like the window borders.
    ExcludeClipRect(DC, R.Left, R.Top, R.Right, R.Bottom);
    // Convert coordinates back into screen-based.
    OffsetClipRgn(DC, WR.Left, WR.Top);
    // Get the region that is now described by the clipping region.
    GetClipRgn(DC, MyRgn);
    // Pass that region on to the default WM_NCPAINT handler.  Remember, we excluded the
    // rectangle that we painted, so Windows will not be able to paint over what we did.
    // Most gradient captions components just let windows draw its stuff first, and then
    // paint the gradient.  This results in an irritating "flicker", caused by the area
    // being painted normally, and then painted over a second time by the gradient.
    Msg.Result := DefWindowProc(Handle, Msg.Msg, MyRgn, Msg.lParam);
  finally
    // If we had to create our own region, we have to clean it up.
    if DeleteRgn then
      DeleteObject(MyRgn);
    ReleaseDC(Handle, DC); // NEVER leave this hanging.
  end;
end;

// Windows sends this message if the user changes any of the system colors.
procedure TGradientForm.WMSysColorChange(var Msg: TWMSysColorChange);
begin
  CalculateColors;
  inherited;
end;

end.

